(window.webpackJsonp=window.webpackJsonp||[]).push([[188],{1065:function(t,r,e){"use strict";e.r(r);var n={asyncData:function(t){var r=t.redirect,e=t.store;if(!e.state.libraries.currentLibraryId)return r("/oops?message=No libraries");r("/library/".concat(e.state.libraries.currentLibraryId))},data:function(){return{}},computed:{streamLibraryItem:function(){return this.$store.state.streamLibraryItem}},methods:{},mounted:function(){},beforeDestroy:function(){}},o=e(4),component=Object(o.a)(n,(function(){return(0,this._self._c)("div",{staticClass:"page",class:this.streamLibraryItem?"streaming":""})}),[],!1,null,null,null);r.default=component.exports}}]);