(window.webpackJsonp=window.webpackJsonp||[]).push([[116],{874:function(t,e,r){"use strict";r.r(e);var l={props:{libraryStats:{type:Object,default:function(){}}},data:function(){return{}},computed:{currentLibraryMediaType:function(){return this.$store.getters["libraries/getCurrentLibraryMediaType"]},isBookLibrary:function(){return"book"===this.currentLibraryMediaType},user:function(){return this.$store.state.user.user},totalItems:function(){var t;return(null===(t=this.libraryStats)||void 0===t?void 0:t.totalItems)||0},totalAuthors:function(){var t;return(null===(t=this.libraryStats)||void 0===t?void 0:t.totalAuthors)||0},numAudioTracks:function(){var t;return(null===(t=this.libraryStats)||void 0===t?void 0:t.numAudioTracks)||0},totalDuration:function(){var t;return(null===(t=this.libraryStats)||void 0===t?void 0:t.totalDuration)||0},totalHours:function(){return Math.round(this.totalDuration/3600)},totalSizePretty:function(){var t,e=(null===(t=this.libraryStats)||void 0===t?void 0:t.totalSize)||0;return this.$bytesPretty(e,1)},totalSizeNum:function(){return this.totalSizePretty.split(" ")[0]},totalSizeMod:function(){return this.totalSizePretty.split(" ")[1]},useOverallHours:function(){return this.totalHours<1e4},totalTime:function(){return this.useOverallHours?this.totalHours:Math.round(this.totalHours/24)}},methods:{},mounted:function(){}},o=r(4),component=Object(o.a)(l,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"flex flex-wrap justify-center mt-6"},[e("div",{staticClass:"flex p-2"},[e("span",{staticClass:"material-symbols text-5xl py-1"},[t._v("newsstand")]),t._v(" "),e("div",{staticClass:"px-1"},[e("p",{staticClass:"text-4.5xl leading-none font-bold"},[t._v(t._s(t.$formatNumber(t.totalItems)))]),t._v(" "),e("p",{staticClass:"text-xs md:text-sm text-white/80"},[t._v(t._s(t.$strings.LabelStatsItemsInLibrary))])])]),t._v(" "),e("div",{staticClass:"flex p-2"},[e("span",{staticClass:"material-symbols text-5xl py-1"},[t._v("show_chart")]),t._v(" "),e("div",{staticClass:"px-1"},[e("p",{staticClass:"text-4.5xl leading-none font-bold"},[t._v(t._s(t.$formatNumber(t.totalTime)))]),t._v(" "),e("p",{staticClass:"text-xs md:text-sm text-white/80"},[t._v(t._s(t.useOverallHours?t.$strings.LabelStatsOverallHours:t.$strings.LabelStatsOverallDays))])])]),t._v(" "),t.isBookLibrary?e("div",{staticClass:"flex p-2"},[e("span",{staticClass:"material-symbols text-5xl py-1"},[t._v("person")]),t._v(" "),e("div",{staticClass:"px-1"},[e("p",{staticClass:"text-4.5xl leading-none font-bold"},[t._v(t._s(t.$formatNumber(t.totalAuthors)))]),t._v(" "),e("p",{staticClass:"text-xs md:text-sm text-white/80"},[t._v(t._s(t.$strings.LabelStatsAuthors))])])]):t._e(),t._v(" "),e("div",{staticClass:"flex p-2"},[e("span",{staticClass:"material-symbols text-5xl pt-1"},[t._v("insert_drive_file")]),t._v(" "),e("div",{staticClass:"px-1"},[e("p",{staticClass:"text-4.5xl leading-none font-bold"},[t._v(t._s(t.$formatNumber(t.totalSizeNum)))]),t._v(" "),e("p",{staticClass:"text-xs md:text-sm text-white/80"},[t._v(t._s(t.$strings.LabelSize)+" ("+t._s(t.totalSizeMod)+")")])])]),t._v(" "),e("div",{staticClass:"flex p-2"},[e("span",{staticClass:"material-symbols text-5xl pt-1"},[t._v("audio_file")]),t._v(" "),e("div",{staticClass:"px-1"},[e("p",{staticClass:"text-4.5xl leading-none font-bold"},[t._v(t._s(t.$formatNumber(t.numAudioTracks)))]),t._v(" "),e("p",{staticClass:"text-xs md:text-sm text-white/80"},[t._v(t._s(t.$strings.LabelStatsAudioTracks))])])])])}),[],!1,null,null,null);e.default=component.exports}}]);