import { ExtractorData } from './js/ExtractorData';
import { getUnrar } from './js/unrar.singleton';
export * from './js/Extractor';
export async function createExtractorFromData({ wasmBinary, data, password = '', }) {
    const unrar = await getUnrar(wasmBinary && { wasmBinary });
    const extractor = new ExtractorData(unrar, data, password);
    unrar.extractor = extractor;
    return extractor;
}
//# sourceMappingURL=index.esm.js.map