export declare const noTrueLogging: () => void;
export declare const noStringOperators: () => void;
export declare const noBoolOperatorAliases: () => void;
export declare const noDoubleNestedGroup: () => void;
export declare const unsupportedEngine: () => void;
