(window.webpackJsonp=window.webpackJsonp||[]).push([[141],{686:function(t,n,e){"use strict";e.r(n);e(83),e(63);var c={props:{accept:{type:String,default:".png, .jpg, .jpeg, .webp"}},data:function(){return{}},computed:{},methods:{reset:function(){this.$refs.fileInput&&(this.$refs.fileInput.value="")},clickUpload:function(){this.$refs.fileInput&&this.$refs.fileInput.click()},inputChanged:function(t){if(t.target&&t.target.files){var n=Array.from(t.target.files);if(n&&n.length){var e=n[0];this.$emit("change",e)}}}},mounted:function(){}},l=e(4),component=Object(l.a)(c,(function(){var t=this,n=t._self._c;return n("div",[n("input",{ref:"fileInput",staticClass:"hidden",attrs:{type:"file",accept:t.accept},on:{change:t.inputChanged}}),t._v(" "),n("ui-btn",{staticClass:"hidden md:block w-full",attrs:{color:"bg-primary",type:"text"},on:{click:t.clickUpload}},[t._t("default")],2),t._v(" "),n("ui-icon-btn",{staticClass:"block md:hidden",attrs:{icon:"upload"},on:{click:t.clickUpload}})],1)}),[],!1,null,null,null);n.default=component.exports;installComponents(component,{UiBtn:e(42).default,UiIconBtn:e(148).default})}}]);