(window.webpackJsonp=window.webpackJsonp||[]).push([[190],{1080:function(e,r,t){"use strict";t.r(r);var n=t(16),o=t(7),c=(t(29),t(26),t(150),t(8),{asyncData:function(e){return Object(o.a)(regeneratorRuntime.mark((function r(){var t,o,c,d,l,f,y,m;return regeneratorRuntime.wrap((function(r){for(;;)switch(r.prev=r.next){case 0:return t=e.params,o=e.query,c=e.store,d=e.redirect,l=t.library,r.next=4,c.dispatch("libraries/fetch",l);case 4:if(f=r.sent){r.next=7;break}return r.abrupt("return",d("/oops?message=Library not found"));case 7:if((o.filter||o.sort||o.desc)&&(y="series"===t.id,m=Object(n.a)(Object(n.a)(Object(n.a)({},y?"seriesFilterBy":"filterBy",o.filter||void 0),y?"seriesSortBy":"orderBy",o.sort||void 0),y?"seriesSortDesc":"orderDesc","0"!=o.desc&&("1"==o.desc||void 0)),c.dispatch("user/updateUserSettings",m)),"podcast"!==f.library.mediaType||"collections"!==t.id&&"series"!==t.id&&"authors"!==t.id){r.next=11;break}return r.abrupt("return",d("/library/".concat(l)));case 11:return r.abrupt("return",{id:t.id||"",libraryId:l});case 12:case"end":return r.stop()}}),r)})))()},data:function(){return{}},computed:{streamLibraryItem:function(){return this.$store.state.streamLibraryItem}},methods:{}}),d=t(4),component=Object(d.a)(c,(function(){var e=this,r=e._self._c;return r("div",{staticClass:"page",class:e.streamLibraryItem?"streaming":""},[r("app-book-shelf-toolbar",{attrs:{page:e.id||""}}),e._v(" "),r("app-lazy-bookshelf",{attrs:{page:e.id||""}})],1)}),[],!1,null,null,null);r.default=component.exports;installComponents(component,{AppBookShelfToolbar:t(690).default,AppLazyBookshelf:t(832).default})}}]);