/* eslint-disable @typescript-eslint/no-explicit-any */
import unrarFactory from './unrar';
let unrar;
export async function getUnrar(options) {
    if (!unrar) {
        unrar = await unrarFactory(options);
    }
    return unrar;
}
//# sourceMappingURL=unrar.singleton.js.map