(window.webpackJsonp=window.webpackJsonp||[]).push([[177,60],{1050:function(t,e,r){"use strict";r.r(e);var n={asyncData:function(t){var e=t.store,r=t.redirect;e.getters["user/getIsAdminOrUp"]||r("/")},data:function(){return{}},watch:{},computed:{},methods:{init:function(){}},mounted:function(){this.init()},beforeDestroy:function(){}},o=r(4),component=Object(o.a)(n,(function(){var t=this,e=t._self._c;return e("div",[e("app-settings-content",{attrs:{"header-text":t.$strings.HeaderItemMetadataUtils}},[e("nuxt-link",{staticClass:"block w-full rounded-sm bg-primary/40 hover:bg-primary/60 text-gray-300 hover:text-white p-4 mt-6 mb-2",attrs:{to:"/config/item-metadata-utils/tags"}},[e("div",{staticClass:"flex justify-between"},[e("p",[t._v(t._s(t.$strings.HeaderManageTags))]),t._v(" "),e("span",{staticClass:"material-symbols"},[t._v("arrow_forward")])])]),t._v(" "),e("nuxt-link",{staticClass:"block w-full rounded-sm bg-primary/40 hover:bg-primary/60 text-gray-300 hover:text-white p-4 my-2",attrs:{to:"/config/item-metadata-utils/genres"}},[e("div",{staticClass:"flex justify-between"},[e("p",[t._v(t._s(t.$strings.HeaderManageGenres))]),t._v(" "),e("span",{staticClass:"material-symbols"},[t._v("arrow_forward")])])]),t._v(" "),e("nuxt-link",{staticClass:"block w-full rounded-sm bg-primary/40 hover:bg-primary/60 text-gray-300 hover:text-white p-4 my-2",attrs:{to:"/config/item-metadata-utils/custom-metadata-providers"}},[e("div",{staticClass:"flex justify-between"},[e("p",[t._v(t._s(t.$strings.HeaderCustomMetadataProviders))]),t._v(" "),e("span",{staticClass:"material-symbols"},[t._v("arrow_forward")])])])],1)],1)}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{AppSettingsContent:r(653).default})},647:function(t,e,r){var content=r(649);content.__esModule&&(content=content.default),"string"==typeof content&&(content=[[t.i,content,""]]),content.locals&&(t.exports=content.locals);(0,r(19).default)("6d988fa2",content,!0,{sourceMap:!1})},648:function(t,e,r){"use strict";r(647)},649:function(t,e,r){var n=r(18)((function(i){return i[1]}));n.push([t.i,"\n#settings-description a {\n  color: rgb(96 165 250);\n}\n#settings-description a:hover {\n  color: rgb(147 197 253);\n  text-decoration-line: underline;\n}\n#settings-description code {\n  font-size: 0.875rem;\n  border-radius: 6px;\n  background-color: rgb(82, 82, 82);\n  color: white;\n  padding: 2px 4px;\n}\n",""]),n.locals={},t.exports=n},653:function(t,e,r){"use strict";r.r(e);r(22),r(43);var n={props:{headerText:String,description:String,note:String},methods:{}},o=(r(648),r(4)),component=Object(o.a)(n,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"bg-bg rounded-md shadow-lg border border-white/5 p-2 sm:p-4 mb-8"},[e("div",{staticClass:"flex items-center mb-2"},[t._t("header-prefix"),t._v(" "),e("h1",{staticClass:"text-xl"},[t._v(t._s(t.headerText))]),t._v(" "),t._t("header-items")],2),t._v(" "),t.description?e("p",{staticClass:"mb-6 text-gray-200",attrs:{id:"settings-description"},domProps:{innerHTML:t._s(t.description)}}):t._e(),t._v(" "),t._t("default")],2)}),[],!1,null,null,null);e.default=component.exports}}]);