(window.webpackJsonp=window.webpackJsonp||[]).push([[150],{658:function(r,e,t){"use strict";t.r(e);var n={props:{type:{type:String,default:"error"}},data:function(){return{}},computed:{icon:function(){return"error"===this.type||"warning"===this.type?"report":"info"},wrapperClass:function(){switch(this.type){case"error":return"bg-error/5 border-error/60 text-error";case"warning":return"bg-warning/5 border-warning/60 text-warning";case"success":return"bg-success/5 border-success/60 text-success";case"info":return"bg-info/5 border-info/60 text-info";default:return"bg-primary/5 border-primary/60 text-primary"}}},methods:{},mounted:function(){}},o=t(4),component=Object(o.a)(n,(function(){var r=this,e=r._self._c;return e("div",{staticClass:"w-full border rounded-lg flex items-center relative py-4 pl-16",class:r.wrapperClass},[e("div",{staticClass:"absolute top-0 left-4 h-full flex items-center"},[e("span",{staticClass:"material-symbols text-2xl"},[r._v(r._s(r.icon))])]),r._v(" "),r._t("default")],2)}),[],!1,null,null,null);e.default=component.exports}}]);