import { ExtractorFile } from './js/ExtractorFile';
import { getUnrar } from './js/unrar.singleton';
export * from './index.esm';
export async function createExtractorFromFile({ wasmBinary, filepath, targetPath = '', password = '', filenameTransform = (filename) => filename, }) {
    const unrar = await getUnrar(wasmBinary && { wasmBinary });
    const extractor = new ExtractorFile(unrar, filepath, targetPath, password, filenameTransform);
    unrar.extractor = extractor;
    return extractor;
}
//# sourceMappingURL=index.js.map