(window.webpackJsonp=window.webpackJsonp||[]).push([[200],{1064:function(e,t,n){"use strict";n.r(t);var c={asyncData:function(e){return{message:e.query.message||""}},data:function(){return{}},computed:{},methods:{},mounted:function(){}},r=n(4),component=Object(r.a)(c,(function(){var e=this,t=e._self._c;return t("div",{staticClass:"w-screen h-screen overflow-hidden page"},[t("div",{staticClass:"flex h-1/3 items-center justify-center"},[t("h1",{staticClass:"text-2xl"},[e._v("Oops... "+e._s(e.message))])])])}),[],!1,null,null,null);t.default=component.exports}}]);