(window.webpackJsonp=window.webpackJsonp||[]).push([[27],{385:function(t){t.exports=JSON.parse('{"ButtonAdd":"افزودن","ButtonAuthors":"ناشر","ButtonBack":"بازگشت","ButtonCancel":"انصراف","ButtonClearFilter":"حذف صافی","ButtonCloseFeed":"بستن فید","ButtonCollections":"مجموعه ها","ButtonCreate":"ساختن","ButtonDelete":"حذف","ButtonHome":"خانه","ButtonIssues":"مشکلات","ButtonLatest":"جدیدترین","ButtonLibrary":"کتابخانه","ButtonOk":"تایید","ButtonOpenFeed":"باز کردن فید","ButtonPause":"توقف","ButtonPlay":"پخش","ButtonPlaylists":"لیست پخش","ButtonRead":"خواندن","ButtonReadLess":"خواندن کمتر","ButtonReadMore":"خواندن بیشتر","ButtonRemove":"حذف","ButtonSave":"ذخیره","ButtonSearch":"جستجو","ButtonSeries":"مجموعه","ButtonSubmit":"ثبت"}')}}]);