(window.webpackJsonp=window.webpackJsonp||[]).push([[158],{692:function(t,l,e){"use strict";e.r(l);var n={props:{value:Boolean},data:function(){return{}},computed:{},methods:{},mounted:function(){}},r=e(4),component=Object(r.a)(n,(function(){var t=this._self._c;return t("div",{staticClass:"w-3 h-3"},[this.value?t("div",{staticClass:"w-full h-full text-sm mr-2 text-success animate-pulse"},[t("svg",{attrs:{viewBox:"0 0 24 24"}},[t("path",{attrs:{fill:"currentColor",d:"M12,2A10,10 0 0,0 2,12A10,10 0 0,0 12,22A10,10 0 0,0 22,12A10,10 0 0,0 12,2Z"}})])]):t("svg",{staticClass:"w-full h-full mr-2 text-white/20",attrs:{viewBox:"0 0 24 24"}},[t("path",{attrs:{fill:"currentColor",d:"M12,2A10,10 0 0,0 2,12A10,10 0 0,0 12,22A10,10 0 0,0 22,12A10,10 0 0,0 12,2Z"}})])])}),[],!1,null,null,null);l.default=component.exports}}]);