(window.webpackJsonp=window.webpackJsonp||[]).push([[77],{838:function(t,e,n){"use strict";n.r(e);n(15);var l={props:{value:Boolean,apiKey:{type:Object,default:function(){return null}}},data:function(){return{}},computed:{show:{get:function(){return this.value},set:function(t){this.$emit("input",t)}},title:function(){return this.$strings.HeaderNewApiKey},apiKeyName:function(){var t;return(null===(t=this.apiKey)||void 0===t?void 0:t.name)||""},apiKeyKey:function(){var t;return(null===(t=this.apiKey)||void 0===t?void 0:t.apiKey)||""}},methods:{},mounted:function(){}},o=n(4),component=Object(o.a)(l,(function(){var t=this,e=t._self._c;return e("modals-modal",{ref:"modal",attrs:{name:"api-key-created",width:800,height:"unset",persistent:""},scopedSlots:t._u([{key:"outer",fn:function(){return[e("div",{staticClass:"absolute top-0 left-0 p-5 w-2/3 overflow-hidden"},[e("p",{staticClass:"text-3xl text-white truncate"},[t._v(t._s(t.title))])])]},proxy:!0}]),model:{value:t.show,callback:function(e){t.show=e},expression:"show"}},[t._v(" "),e("form",{on:{submit:function(e){return e.preventDefault(),t.submitForm.apply(null,arguments)}}},[e("div",{staticClass:"px-4 w-full text-sm py-6 rounded-lg bg-bg shadow-lg border border-black-300 overflow-y-auto overflow-x-hidden",staticStyle:{"min-height":"200px","max-height":"80vh"}},[e("div",{staticClass:"w-full p-8"},[e("p",{staticClass:"text-lg text-white mb-4"},[t._v(t._s(t.$getString("LabelApiKeyCreated",[t.apiKeyName])))]),t._v(" "),e("p",{staticClass:"text-lg text-white mb-4"},[t._v(t._s(t.$strings.LabelApiKeyCreatedDescription))]),t._v(" "),e("ui-text-input",{attrs:{label:"API Key",value:t.apiKeyKey,readonly:"","show-copy":""}}),t._v(" "),e("div",{staticClass:"flex justify-end mt-4"},[e("ui-btn",{attrs:{color:"bg-primary"},on:{click:function(e){t.show=!1}}},[t._v(t._s(t.$strings.ButtonClose))])],1)],1)])])])}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{UiTextInput:n(68).default,UiBtn:n(42).default,ModalsModal:n(39).default})}}]);