(window.webpackJsonp=window.webpackJsonp||[]).push([[191],{1069:function(e,t,r){"use strict";r.r(t);var n=r(7),o=(r(29),{asyncData:function(e){return Object(n.a)(regeneratorRuntime.mark((function t(){var r,n,o,c,l;return regeneratorRuntime.wrap((function(t){for(;;)switch(t.prev=t.next){case 0:return r=e.store,n=e.params,o=e.redirect,c=n.library,t.next=4,r.dispatch("libraries/fetch",c);case 4:if(l=t.sent){t.next=7;break}return t.abrupt("return",o('/oops?message=Library "'.concat(c,'" not found')));case 7:return t.abrupt("return",{library:l});case 8:case"end":return t.stop()}}),t)})))()},data:function(){return{}},computed:{streamLibraryItem:function(){return this.$store.state.streamLibraryItem}},methods:{},mounted:function(){},beforeDestroy:function(){}}),c=r(4),component=Object(c.a)(o,(function(){var e=this,t=e._self._c;return t("div",{staticClass:"page",class:e.streamLibraryItem?"streaming":""},[t("app-book-shelf-toolbar",{attrs:{"is-home":""}}),e._v(" "),t("app-book-shelf-categorized")],1)}),[],!1,null,null,null);t.default=component.exports;installComponents(component,{AppBookShelfToolbar:r(690).default,AppBookShelfCategorized:r(866).default})}}]);