(window.webpackJsonp=window.webpackJsonp||[]).push([[126],{717:function(t,e,r){var content=r(767);content.__esModule&&(content=content.default),"string"==typeof content&&(content=[[t.i,content,""]]),content.locals&&(t.exports=content.locals);(0,r(19).default)("4c9c9f37",content,!0,{sourceMap:!1})},766:function(t,e,r){"use strict";r(717)},767:function(t,e,r){var o=r(18)((function(i){return i[1]}));o.push([t.i,"\n#providers {\n  table-layout: fixed;\n  border-collapse: collapse;\n  border: 1px solid #474747;\n  width: 100%;\n}\n#providers td,\n#providers th {\n  /* border: 1px solid #2e2e2e; */\n  padding: 8px 8px;\n  text-align: left;\n}\n#providers td.py-0 {\n  padding: 0px 8px;\n}\n#providers tr:nth-child(even) {\n  background-color: #373838;\n}\n#providers tr:nth-child(odd) {\n  background-color: #2f2f2f;\n}\n#providers tr:hover {\n  background-color: #444;\n}\n#providers th {\n  font-size: 0.8rem;\n  font-weight: 600;\n  padding-top: 5px;\n  padding-bottom: 5px;\n  background-color: #272727;\n}\n.custom-provider-api-key {\n  padding: 1px;\n  background-color: #272727;\n  border-radius: 4px;\n  color: transparent;\n  transition: color, background-color 0.5s ease;\n}\n.custom-provider-api-key:hover {\n  background-color: transparent;\n  color: white;\n}\n",""]),o.locals={},t.exports=o},852:function(t,e,r){"use strict";r.r(e);r(15),r(8);var o={props:{providers:{type:Array,default:function(){return[]}},processing:Boolean},data:function(){return{}},methods:{removeProvider:function(t){var e=this,r={message:this.$getString("MessageConfirmDeleteMetadataProvider",[t.name]),callback:function(r){r&&(e.$emit("update:processing",!0),e.$axios.$delete("/api/custom-metadata-providers/".concat(t.id)).then((function(){e.$toast.success(e.$strings.ToastProviderRemoveSuccess),e.$emit("removed",t.id)})).catch((function(t){console.error("Failed to remove provider",t),e.$toast.error(e.$strings.ToastRemoveFailed)})).finally((function(){e.$emit("update:processing",!1)})))},type:"yesNo"};this.$store.commit("globals/setConfirmPrompt",r)}}},n=(r(766),r(4)),component=Object(n.a)(o,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"min-h-40"},[t.providers.length?e("table",{attrs:{id:"providers"}},[e("tr",[e("th",[t._v(t._s(t.$strings.LabelName))]),t._v(" "),e("th",[t._v("URL")]),t._v(" "),e("th",[t._v("Authorization Header Value")]),t._v(" "),e("th",{staticClass:"w-12"})]),t._v(" "),t._l(t.providers,(function(r){return e("tr",{key:r.id},[e("td",{staticClass:"text-sm"},[t._v(t._s(r.name))]),t._v(" "),e("td",{staticClass:"text-sm"},[t._v(t._s(r.url))]),t._v(" "),e("td",{staticClass:"text-sm"},[r.authHeaderValue?e("span",{staticClass:"custom-provider-api-key"},[t._v(t._s(r.authHeaderValue))]):t._e()]),t._v(" "),e("td",{staticClass:"py-0"},[e("div",{staticClass:"h-8 w-8 flex items-center justify-center text-white/50 hover:text-error cursor-pointer",on:{click:function(e){return e.stopPropagation(),t.removeProvider(r)}}},[e("button",{staticClass:"material-symbols text-base",attrs:{type:"button","aria-label":t.$strings.ButtonDelete}},[t._v("delete")])])])])}))],2):t.processing?t._e():e("div",{staticClass:"text-center py-8"},[e("p",{staticClass:"text-lg"},[t._v(t._s(t.$strings.LabelNoCustomMetadataProviders))])]),t._v(" "),t.processing?e("div",{staticClass:"absolute inset-0 h-full flex items-center justify-center bg-black/40 rounded-md"},[e("ui-loading-indicator")],1):t._e()])}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{UiLoadingIndicator:r(348).default})}}]);