(window.webpackJsonp=window.webpackJsonp||[]).push([[70],{1082:function(t,e,r){"use strict";r.r(e);r(22),r(43),r(17);var n={props:{feed:{type:Object,default:function(){}},libraryFolderPath:String},data:function(){return{width:900}},computed:{title:function(){return this.metadata.title||"No Title"},image:function(){return this.metadata.imageUrl},description:function(){return this.metadata.description||""},author:function(){return this.metadata.author||""},metadata:function(){return this.feed||{}},numEpisodes:function(){return this.feed.numEpisodes||0},folderPath:function(){return this.libraryFolderPath?"".concat(this.libraryFolderPath,"/").concat(this.$sanitizeFilename(this.title)):""},detailsWidth:function(){return this.width-85}},methods:{},updated:function(){this.width=this.$refs.wrapper.clientWidth},mounted:function(){this.width=this.$refs.wrapper.clientWidth}},l=r(4),component=Object(l.a)(n,(function(){var t=this,e=t._self._c;return e("div",{ref:"wrapper",staticClass:"w-full p-2 border border-white/10 rounded-sm"},[e("div",{staticClass:"flex"},[e("div",{staticClass:"w-16 min-w-16"},[e("div",{staticClass:"w-full h-16 bg-primary"},[t.image?e("img",{staticClass:"w-full h-full object-cover",attrs:{src:t.image}}):t._e()]),t._v(" "),e("p",{staticClass:"text-gray-400 text-xxs pt-1 text-center"},[t._v(t._s(t.numEpisodes)+" "+t._s(t.$strings.HeaderEpisodes))])]),t._v(" "),e("div",{staticClass:"grow pl-2",style:{maxWidth:t.detailsWidth+"px"}},[e("p",{staticClass:"mb-1"},[t._v(t._s(t.title))]),t._v(" "),e("p",{staticClass:"text-xs mb-1 text-gray-300"},[t._v(t._s(t.author))]),t._v(" "),e("p",{staticClass:"text-xs mb-2 text-gray-200"},[t._v(t._s(t.description))]),t._v(" "),e("p",{staticClass:"text-xs truncate text-blue-200"},[t._v("\n        "+t._s(t.$strings.LabelFolder)+": "),e("span",{staticClass:"font-mono"},[t._v(t._s(t.folderPath))])])])])])}),[],!1,null,null,null);e.default=component.exports}}]);