(window.webpackJsonp=window.webpackJsonp||[]).push([[197],{1081:function(e,r,t){"use strict";t.r(r);var n=t(7),o=(t(29),t(17),{asyncData:function(e){return Object(n.a)(regeneratorRuntime.mark((function r(){var t,n,o,c,d,f,l,h;return regeneratorRuntime.wrap((function(r){for(;;)switch(r.prev=r.next){case 0:return t=e.store,n=e.params,o=e.redirect,e.query,c=e.app,d=n.library,r.next=4,t.dispatch("libraries/fetch",d);case 4:if(f=r.sent){r.next=7;break}return r.abrupt("return",o("/oops?message=Library not found"));case 7:if("podcast"!==(l=f.library).mediaType){r.next=10;break}return r.abrupt("return",o("/library/".concat(d)));case 10:return r.next=12,c.$axios.$get("/api/libraries/".concat(l.id,"/series/").concat(n.id,"?include=progress,rssfeed")).catch((function(e){return console.error("Failed",e),!1}));case 12:if(h=r.sent){r.next=15;break}return r.abrupt("return",o("/oops?message=Series not found"));case 15:return r.abrupt("return",{series:h,seriesId:n.id});case 16:case"end":return r.stop()}}),r)})))()},data:function(){return{}},computed:{streamLibraryItem:function(){return this.$store.state.streamLibraryItem}},methods:{seriesUpdated:function(e){this.series=e}},mounted:function(){this.$root.socket&&this.$root.socket.on("series_updated",this.seriesUpdated)},beforeDestroy:function(){this.$root.socket&&this.$root.socket.off("series_updated",this.seriesUpdated)}}),c=t(4),component=Object(c.a)(o,(function(){var e=this,r=e._self._c;return r("div",{staticClass:"page",class:e.streamLibraryItem?"streaming":""},[r("app-book-shelf-toolbar",{attrs:{"selected-series":e.series}}),e._v(" "),r("app-lazy-bookshelf",{attrs:{page:"series-books","series-id":e.seriesId}})],1)}),[],!1,null,null,null);r.default=component.exports;installComponents(component,{AppBookShelfToolbar:t(690).default,AppLazyBookshelf:t(832).default})}}]);