(window.webpackJsonp=window.webpackJsonp||[]).push([[107],{715:function(e,t,n){var content=n(763);content.__esModule&&(content=content.default),"string"==typeof content&&(content=[[e.i,content,""]]),content.locals&&(e.exports=content.locals);(0,n(19).default)("6cf297f0",content,!0,{sourceMap:!1})},762:function(e,t,n){"use strict";n(715)},763:function(e,t,n){var l=n(18)((function(i){return i[1]}));l.push([e.i,"\n.episodesTable[data-v-927273f2] {\n  width: 100%;\n  max-width: 100%;\n  border: 1px solid #474747;\n  display: flex;\n  flex-direction: column;\n}\n.episodesTable div.header[data-v-927273f2] {\n  background-color: #272727;\n  display: flex;\n  flex-direction: column;\n  justify-content: center;\n  align-items: flex-start;\n  padding: 4px 8px;\n}\n.episodesTable .scroller[data-v-927273f2] {\n  display: flex;\n  flex-direction: column;\n  max-height: 250px;\n  overflow-x: hidden;\n  overflow-y: scroll;\n}\n.episodesTable .scroller div[data-v-927273f2] {\n  background-color: #373838;\n  padding: 4px 8px;\n  display: flex;\n  align-items: center;\n  justify-content: flex-start;\n  height: 32px;\n  flex: 0 0 32px;\n}\n.episodesTable .scroller div[data-v-927273f2]:nth-child(even) {\n  background-color: #2f2f2f;\n}\n",""]),l.locals={},e.exports=l},845:function(e,t,n){"use strict";n.r(t);n(17);var l={props:{value:Boolean,feed:{type:Object,default:function(){}}},data:function(){return{processing:!1}},computed:{show:{get:function(){return this.value},set:function(e){this.$emit("input",e)}},_feed:function(){return this.feed||{}},feedUrl:function(){return this.feed?"".concat(window.origin).concat(this.$config.routerBasePath).concat(this.feed.feedUrl):""}}},d=(n(762),n(4)),component=Object(d.a)(l,(function(){var e=this,t=e._self._c;return t("modals-modal",{attrs:{name:"rss-feed-view-modal",processing:e.processing,width:700,height:"unset"},model:{value:e.show,callback:function(t){e.show=t},expression:"show"}},[t("div",{ref:"wrapper",staticClass:"px-8 py-6 w-full text-sm rounded-lg bg-bg shadow-lg border border-black-300 relative overflow-hidden"},[e.feed?t("div",{staticClass:"w-full"},[t("p",{staticClass:"text-lg font-semibold mb-4"},[e._v(e._s(e.$strings.HeaderRSSFeedGeneral))]),e._v(" "),t("div",{staticClass:"w-full relative"},[t("ui-text-input",{attrs:{value:e.feedUrl,readonly:"","show-copy":""}})],1),e._v(" "),e.feed.meta?t("div",{staticClass:"mt-5"},[t("div",{staticClass:"flex py-0.5"},[t("div",{staticClass:"w-48"},[t("span",{staticClass:"text-white/60 uppercase text-sm"},[e._v(e._s(e.$strings.LabelRSSFeedPreventIndexing))])]),e._v(" "),t("div",[e._v(e._s(e.feed.meta.preventIndexing?"Yes":"No"))])]),e._v(" "),e.feed.meta.ownerName?t("div",{staticClass:"flex py-0.5"},[t("div",{staticClass:"w-48"},[t("span",{staticClass:"text-white/60 uppercase text-sm"},[e._v(e._s(e.$strings.LabelRSSFeedCustomOwnerName))])]),e._v(" "),t("div",[e._v(e._s(e.feed.meta.ownerName))])]):e._e(),e._v(" "),e.feed.meta.ownerEmail?t("div",{staticClass:"flex py-0.5"},[t("div",{staticClass:"w-48"},[t("span",{staticClass:"text-white/60 uppercase text-sm"},[e._v(e._s(e.$strings.LabelRSSFeedCustomOwnerEmail))])]),e._v(" "),t("div",[e._v(e._s(e.feed.meta.ownerEmail))])]):e._e()]):e._e(),e._v(" "),t("div",{staticClass:"episodesTable mt-2"},[t("div",{staticClass:"bg-primary/40 h-12 header"},[e._v("\n          "+e._s(e.$strings.LabelEpisodeTitle)+"\n        ")]),e._v(" "),t("div",{staticClass:"scroller"},e._l(e.feed.episodes,(function(n){return t("div",{key:n.id,staticClass:"h-8 text-xs truncate"},[e._v("\n            "+e._s(n.title)+"\n          ")])})),0)])]):e._e()])])}),[],!1,null,"927273f2",null);t.default=component.exports;installComponents(component,{UiTextInput:n(68).default,ModalsModal:n(39).default})}}]);