(window.webpackJsonp=window.webpackJsonp||[]).push([[72],{860:function(t,e,n){"use strict";n.r(e);n(33),n(8);var l={props:{value:String,items:{type:Array,default:function(){return[]}}},data:function(){return{showMenu:!1}},computed:{selected:{get:function(){return this.value},set:function(t){this.$emit("input",t)}},selectedText:function(){var t=this;if(!this.selected)return"";var filter=this.items.find((function(i){return i.value===t.selected}));return filter?filter.text:""},filterData:function(){return this.$store.state.libraries.filterData||{}}},methods:{clearSelected:function(){var t=this;this.selected="all",this.showMenu=!1,this.$nextTick((function(){return t.$emit("change","all")}))},clickOutside:function(){this.showMenu=!1},clickedOption:function(option){var t=this,e=option.value;this.selected!==e?(this.selected=e,this.showMenu=!1,this.$nextTick((function(){return t.$emit("change",e)}))):this.showMenu=!1}}},r=n(4),component=Object(r.a)(l,(function(){var t=this,e=t._self._c;return e("div",{directives:[{name:"click-outside",rawName:"v-click-outside",value:t.clickOutside,expression:"clickOutside"}],ref:"wrapper",staticClass:"relative"},[e("div",{staticClass:"relative h-9"},[e("button",{staticClass:"relative w-full h-full border border-gray-500 hover:border-gray-400 rounded-sm shadow-xs pl-3 pr-3 py-0 text-left focus:outline-hidden cursor-pointer",attrs:{type:"button","aria-haspopup":"menu","aria-expanded":t.showMenu},on:{click:function(e){e.preventDefault(),t.showMenu=!t.showMenu}}},[e("span",{staticClass:"flex items-center justify-between"},[e("span",{staticClass:"block truncate text-xs"},[t._v(t._s(t.selectedText))])])]),t._v(" "),"all"===t.selected?e("span",{staticClass:"ml-3 absolute inset-y-0 right-0 flex items-center pr-2 pointer-events-none"},[e("svg",{staticClass:"h-5 w-5 text-gray-400",attrs:{xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor","aria-hidden":"true"}},[e("path",{attrs:{"fill-rule":"evenodd",d:"M10 3a1 1 0 01.707.293l3 3a1 1 0 01-1.414 1.414L10 5.414 7.707 7.707a1 1 0 01-1.414-1.414l3-3A1 1 0 0110 3zm-3.707 9.293a1 1 0 011.414 0L10 14.586l2.293-2.293a1 1 0 011.414 1.414l-3 3a1 1 0 01-1.414 0l-3-3a1 1 0 010-1.414z","clip-rule":"evenodd"}})])]):e("button",{staticClass:"ml-3 absolute inset-y-0 right-0 flex items-center pr-2 cursor-pointer text-gray-400 hover:text-gray-300",attrs:{type:"button","aria-label":t.$strings.ButtonClearFilter},on:{mousedown:function(t){t.stopPropagation()},mouseup:function(t){t.stopPropagation()},click:function(e){return e.stopPropagation(),e.preventDefault(),t.clearSelected.apply(null,arguments)}}},[e("span",{staticClass:"material-symbols",staticStyle:{"font-size":"1.1rem"}},[t._v("close")])])]),t._v(" "),e("div",{directives:[{name:"show",rawName:"v-show",value:t.showMenu,expression:"showMenu"}],staticClass:"absolute z-10 mt-1 w-full bg-bg border border-black-200 shadow-lg max-h-96 rounded-md py-1 text-sm ring-1 ring-black/5 overflow-auto focus:outline-hidden"},[e("ul",{staticClass:"h-full w-full",attrs:{role:"menu"}},[t._l(t.items,(function(n){return[e("li",{key:n.value,staticClass:"select-none relative py-2 pr-9 cursor-pointer hover:bg-white/5",class:n.value===t.selected?"bg-white/5 text-yellow-400":"text-gray-200 hover:text-white",attrs:{role:"menuitem"},on:{click:function(e){return t.clickedOption(n)}}},[e("div",{staticClass:"flex items-center justify-between"},[e("span",{staticClass:"font-normal ml-3 block truncate"},[t._v(t._s(n.text))])]),t._v(" "),n.value===t.selected?e("div",{staticClass:"absolute inset-y-0 right-2 h-full flex items-center pointer-events-none"},[e("span",{staticClass:"material-symbols text-base text-yellow-400"},[t._v("check")])]):t._e()])]}))],2)])])}),[],!1,null,null,null);e.default=component.exports}}]);