(window.webpackJsonp=window.webpackJsonp||[]).push([[68],{870:function(t,e,r){"use strict";r.r(e);r(17),r(15),r(14);var n={props:{narrator:{type:Object,default:function(){}},width:Number,height:{type:Number,default:100}},data:function(){return{}},computed:{cardWidth:function(){return 1.5*this.cardHeight},cardHeight:function(){return this.height*this.sizeMultiplier},name:function(){var t;return(null===(t=this.narrator)||void 0===t?void 0:t.name)||""},numBooks:function(){var t,e;return(null===(t=this.narrator)||void 0===t?void 0:t.numBooks)||(null===(e=this.narrator)||void 0===e||null===(e=e.books)||void 0===e?void 0:e.length)||0},userCanUpdate:function(){return this.$store.getters["user/getUserCanUpdate"]},currentLibraryId:function(){return this.$store.state.libraries.currentLibraryId},sizeMultiplier:function(){return this.$store.getters["user/getSizeMultiplier"]}},methods:{}},o=r(4),component=Object(o.a)(n,(function(){var t=this,e=t._self._c;return e("div",[e("nuxt-link",{attrs:{to:"/library/".concat(t.currentLibraryId,"/bookshelf?filter=narrators.").concat(t.$encode(t.name))}},[e("div",{staticClass:"bg-primary box-shadow-book rounded-md relative overflow-hidden",style:{width:t.cardWidth+"px",height:t.cardHeight+"px",fontSize:t.sizeMultiplier+"rem"},attrs:{"cy-id":"card"}},[e("div",{staticClass:"absolute inset-0 w-full h-full flex items-center justify-center pointer-events-none opacity-40"},[e("span",{staticClass:"material-symbols text-[10em]"},[t._v("")])]),t._v(" "),e("div",{staticClass:"absolute bottom-0 left-0 w-full py-1e bg-black/60 px-2e"},[e("p",{staticClass:"text-center font-semibold truncate text-gray-200",style:{fontSize:"0.75em"},attrs:{"cy-id":"name"}},[t._v(t._s(t.name))]),t._v(" "),e("p",{staticClass:"text-center text-gray-200",style:{fontSize:"0.65em"},attrs:{"cy-id":"numBooks"}},[t._v(t._s(t.numBooks)+" Book"+t._s(1===t.numBooks?"":"s"))])])])])],1)}),[],!1,null,null,null);e.default=component.exports}}]);