(window.webpackJsonp=window.webpackJsonp||[]).push([[63],{869:function(t,e,r){"use strict";r.r(e);r(17),r(26),r(27),r(15),r(14),r(8);var o={props:{group:{type:Object,default:function(){return null}},width:Number,height:{type:Number,default:192}},data:function(){return{isHovering:!1}},computed:{bookCoverAspectRatio:function(){return this.$store.getters["libraries/getBookCoverAspectRatio"]},cardWidth:function(){return this.width||2*this.cardHeight},cardHeight:function(){return this.height*this.sizeMultiplier},currentLibraryId:function(){return this.$store.state.libraries.currentLibraryId},_group:function(){return this.group||{}},groupType:function(){return this._group.type},groupTo:function(){return"/library/".concat(this.currentLibraryId,"/bookshelf?filter=").concat(this.filter)},sizeMultiplier:function(){return this.$store.getters["user/getSizeMultiplier"]},bookItems:function(){return this._group.books||[]},groupName:function(){return this._group.name||"No Name"},groupEncode:function(){return this.$encode(this.groupName)},filter:function(){return"".concat(this.groupType,".").concat(this.$encode(this.groupName))},hasValidCovers:function(){return!!this.bookItems.map((function(t){return t.media.coverPath})).length}},methods:{mouseoverCard:function(){this.isHovering=!0},mouseleaveCard:function(){this.isHovering=!1},clickCard:function(){this.$emit("click",this.group)}}},n=r(4),component=Object(n.a)(o,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"relative"},[e("div",{staticClass:"rounded-xs h-full relative",style:{width:t.cardWidth+"px",height:t.cardHeight+"px"},on:{mouseover:t.mouseoverCard,mouseleave:t.mouseleaveCard,click:t.clickCard}},[e("nuxt-link",{staticClass:"cursor-pointer",attrs:{to:t.groupTo}},[e("div",{staticClass:"w-full h-full relative",class:t.isHovering?"bg-black-400":"bg-primary"},[e("covers-group-cover",{ref:"groupcover",attrs:{id:t.groupEncode,name:t.groupName,type:t.groupType,"book-items":t.bookItems,width:t.cardWidth,height:t.cardHeight,"book-cover-aspect-ratio":t.bookCoverAspectRatio}}),t._v(" "),t.hasValidCovers?e("div",{staticClass:"bg-black/60 absolute top-0 left-0 w-full h-full flex items-center justify-center text-center transition-opacity z-30",class:t.isHovering?"":"opacity-0",style:{padding:"".concat(t.sizeMultiplier,"rem")}},[e("p",{style:{fontSize:1.2*t.sizeMultiplier+"rem"}},[t._v(t._s(t.groupName))])]):t._e(),t._v(" "),e("div",{staticClass:"absolute z-10 top-1.5e right-1.5e rounded-md leading-3e p-1e font-semibold text-white flex items-center justify-center",staticStyle:{"background-color":"#cd9d49dd"},style:{fontSize:"0.8em"}},[t._v(t._s(t.bookItems.length))])],1)])],1)])}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{CoversGroupCover:r(359).default})}}]);