const globals = {
  SupportedImageTypes: ['png', 'jpg', 'jpeg', 'webp'],
  SupportedAudioTypes: ['m4b', 'mp3', 'm4a', 'flac', 'opus', 'ogg', 'oga', 'mp4', 'aac', 'wma', 'aiff', 'aif', 'wav', 'webm', 'webma', 'mka', 'awb', 'caf', 'mpg', 'mpeg'],
  SupportedEbookTypes: ['epub', 'pdf', 'mobi', 'azw3', 'cbr', 'cbz'],
  TextFileTypes: ['txt', 'nfo'],
  MetadataFileTypes: ['opf', 'abs', 'xml', 'json']
}

module.exports = globals
