(window.webpackJsonp=window.webpackJsonp||[]).push([[147],{680:function(t,e,n){"use strict";n.r(e);var l={props:{value:String,label:String,disabled:{type:Boolean,default:!1}},data:function(){return{}},computed:{content:{get:function(){return this.value},set:function(t){this.$emit("input",t)}}},methods:{trixFileAccept:function(t){t.preventDefault()},blur:function(){this.$refs.input&&this.$refs.input.blur&&this.$refs.input.blur()}},mounted:function(){},beforeDestroy:function(){}},o=n(4),component=Object(o.a)(l,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"default-style"},[t.label?e("p",{staticClass:"px-1 text-sm font-semibold",class:{"text-gray-400":t.disabled},staticStyle:{"margin-top":"0","margin-bottom":"0.125em"}},[t._v("\n    "+t._s(t.label)+"\n  ")]):t._e(),t._v(" "),e("ui-vue-trix",{ref:"input",attrs:{"disabled-editor":t.disabled},on:{"trix-file-accept":t.trixFileAccept},model:{value:t.content,callback:function(e){t.content=e},expression:"content"}})],1)}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{UiVueTrix:n(703).default})}}]);