(window.webpackJsonp=window.webpackJsonp||[]).push([[80],{709:function(t,e,l){"use strict";l.r(e);var o={props:{value:Boolean,title:String,items:{type:Array,default:function(){return[]}},selected:String},data:function(){return{}},computed:{show:{get:function(){return this.value},set:function(t){this.$emit("input",t)}}},methods:{clickedOption:function(t){this.$emit("action",{action:t})}},mounted:function(){}},n=l(4),component=Object(n.a)(o,(function(){var t=this,e=t._self._c;return e("modals-modal",{attrs:{width:300,height:"100%"},scopedSlots:t._u([{key:"outer",fn:function(){return[t.title?e("div",{staticClass:"absolute top-7 left-4 z-40",staticStyle:{"max-width":"80%"}},[e("p",{staticClass:"text-white text-lg truncate"},[t._v(t._s(t.title))])]):t._e()]},proxy:!0}]),model:{value:t.show,callback:function(e){t.show=e},expression:"show"}},[t._v(" "),e("div",{staticClass:"w-full h-full overflow-hidden absolute top-0 left-0 flex items-center justify-center",on:{click:function(e){t.show=!1}}},[e("div",{ref:"container",staticClass:"w-full overflow-x-hidden overflow-y-auto bg-primary rounded-lg border border-white/20",staticStyle:{"max-height":"75%"},on:{click:function(t){t.stopPropagation()}}},[e("ul",{staticClass:"h-full w-full",attrs:{role:"listbox","aria-labelledby":"listbox-label"}},[t._l(t.items,(function(l){return[e("li",{key:l.value,staticClass:"text-gray-50 select-none relative py-4 cursor-pointer hover:bg-black-400",class:t.selected===l.value?"bg-success/10":"",attrs:{role:"option"},on:{click:function(e){return t.clickedOption(l.value)}}},[e("div",{staticClass:"relative flex items-center px-3"},[e("p",{staticClass:"font-normal block truncate text-base text-white/80"},[t._v(t._s(l.text))])])])]}))],2)])])])}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{ModalsModal:l(39).default})}}]);