(window.webpackJsonp=window.webpackJsonp||[]).push([[60],{647:function(t,e,n){var content=n(649);content.__esModule&&(content=content.default),"string"==typeof content&&(content=[[t.i,content,""]]),content.locals&&(t.exports=content.locals);(0,n(19).default)("6d988fa2",content,!0,{sourceMap:!1})},648:function(t,e,n){"use strict";n(647)},649:function(t,e,n){var r=n(18)((function(i){return i[1]}));r.push([t.i,"\n#settings-description a {\n  color: rgb(96 165 250);\n}\n#settings-description a:hover {\n  color: rgb(147 197 253);\n  text-decoration-line: underline;\n}\n#settings-description code {\n  font-size: 0.875rem;\n  border-radius: 6px;\n  background-color: rgb(82, 82, 82);\n  color: white;\n  padding: 2px 4px;\n}\n",""]),r.locals={},t.exports=r},653:function(t,e,n){"use strict";n.r(e);n(22),n(43);var r={props:{headerText:String,description:String,note:String},methods:{}},o=(n(648),n(4)),component=Object(o.a)(r,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"bg-bg rounded-md shadow-lg border border-white/5 p-2 sm:p-4 mb-8"},[e("div",{staticClass:"flex items-center mb-2"},[t._t("header-prefix"),t._v(" "),e("h1",{staticClass:"text-xl"},[t._v(t._s(t.headerText))]),t._v(" "),t._t("header-items")],2),t._v(" "),t.description?e("p",{staticClass:"mb-6 text-gray-200",attrs:{id:"settings-description"},domProps:{innerHTML:t._s(t.description)}}):t._e(),t._v(" "),t._t("default")],2)}),[],!1,null,null,null);e.default=component.exports}}]);