"use strict";
const QueryTypes = module.exports = {
  SELECT: "SELECT",
  INSERT: "INSERT",
  UPDATE: "UPDATE",
  BULKUPDATE: "BULKUPDATE",
  BULKDELETE: "BULKDELETE",
  DELETE: "DELETE",
  UPSERT: "UPSERT",
  VERSION: "VERSION",
  SHOWTABLES: "SHOWTABLES",
  SHOWINDEXES: "SHOWINDEXES",
  DESCRIBE: "DESCRIBE",
  RAW: "RAW",
  FOREIGNKEYS: "FOREIGNKEYS",
  SHOWCONSTRAINTS: "SHOWCONSTRAINTS"
};
//# sourceMappingURL=query-types.js.map
