(window.webpackJsonp=window.webpackJsonp||[]).push([[160],{678:function(t,e,l){"use strict";l.r(e);var o={props:{type:{type:String,default:"full"}},data:function(){return{}},computed:{},methods:{},mounted:function(){}},n=l(4),component=Object(n.a)(o,(function(){var t=this,e=t._self._c;return e("div",["bonus"==t.type?[e("ui-tooltip",{attrs:{text:"Bonus",direction:"top"}},[e("span",{staticClass:"material-symbols ml-1",staticStyle:{"font-size":"0.8rem"}},[t._v("local_play")])])]:t._e(),t._v(" "),"trailer"==t.type?[e("ui-tooltip",{attrs:{text:"Trailer",direction:"top"}},[e("span",{staticClass:"material-symbols ml-1",staticStyle:{"font-size":"0.8rem"}},[t._v("local_movies")])])]:t._e()],2)}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{UiTooltip:l(76).default})}}]);