(window.webpackJsonp=window.webpackJsonp||[]).push([[32],{390:function(t){t.exports=JSON.parse('{"ButtonAdd":"जोड़ें","ButtonAddApiKey":"एपीआई कुंजी जोड़ें","ButtonAddChapters":"अध्याय जोड़ें","ButtonAddDevice":"उपकरण जोड़ें","ButtonAddLibrary":"संग्रह जोड़ें","ButtonAddPodcasts":"पॉडकास्ट जोड़ें","ButtonAddUser":"उपयोगकर्ता जोड़ें","ButtonAddYourFirstLibrary":"अपनी पहली पुस्तकालय जोड़ें","ButtonApply":"लागू करें","ButtonApplyChapters":"अध्यायों में परिवर्तन लागू करें","ButtonAuthors":"लेखक","ButtonBack":"पीछे","ButtonBatchEditPopulateFromExisting":"मौजूदा से आबाद करें","ButtonBatchEditPopulateMapDetails":"मानचित्र विवरण भरें","ButtonBrowseForFolder":"फ़ोल्डर खोजें","ButtonCancel":"रद्द करें","ButtonCancelEncode":"एनकोड रद्द करें","ButtonChangeRootPassword":"रूट का पासवर्ड बदलें","ButtonCheckAndDownloadNewEpisodes":"नए एपिसोड खोजें और डाउनलोड करें","ButtonChooseAFolder":"एक फ़ोल्डर चुनें","ButtonChooseFiles":"फ़ाइलें चुनें","ButtonClearFilter":"लागू फ़िल्टर साफ़ करें","ButtonClose":"बंद करें","ButtonCloseFeed":"फ़ीड बंद करें","ButtonCloseSession":"वर्तमान सत्र बंद करें","ButtonCollections":"संग्रह","ButtonConfigureScanner":"स्कैनर सेटिंग्स बदलें","ButtonCreate":"बनाएं","ButtonCreateBackup":"बैकअप लें","ButtonDelete":"हटाएं","ButtonDownloadQueue":"कतार डाउनलोड करें","ButtonEdit":"संपादित करें","ButtonEditChapters":"अध्याय संपादित करें","ButtonEditPodcast":"पॉडकास्ट संपादित करें","ButtonEnable":"सक्षम करें","ButtonForceReScan":"बलपूर्वक पुन: स्कैन करें","ButtonFullPath":"पूर्ण पथ","ButtonHide":"छुपाएं","ButtonHome":"घर","ButtonIssues":"समस्याएं","ButtonLatest":"नवीनतम","ButtonLibrary":"पुस्तकालय","ButtonLogout":"लॉग आउट","ButtonLookup":"तलाश करें","ButtonManageTracks":"ट्रैक्स मैनेज करें","ButtonMapChapterTitles":"अध्यायों का मिलान करें","ButtonMatchAllAuthors":"सभी लेखकों को तलाश करें","ButtonMatchBooks":"संबंधित पुस्तकों का मिलान करें","ButtonNevermind":"कोई बात नहीं","ButtonOk":"ठीक है","ButtonOpenFeed":"फ़ीड खोलें","ButtonOpenManager":"मैनेजर खोलें","ButtonPlay":"चलाएँ","ButtonPlaying":"चल रही है","ButtonPlaylists":"प्लेलिस्ट्स","ButtonPurgeAllCache":"सभी Cache मिटाएं","ButtonPurgeItemsCache":"आइटम Cache मिटाएं","ButtonQueueAddItem":"क़तार में जोड़ें","ButtonQueueRemoveItem":"कतार से हटाएं","ButtonQuickMatch":"जल्दी से समानता की तलाश करें","ButtonReScan":"पुन: स्कैन करें","ButtonRead":"पढ़ लिया","ButtonRemove":"हटाएं","ButtonRemoveAll":"सभी हटाएं","ButtonRemoveAllLibraryItems":"पुस्तकालय की सभी आइटम हटाएं","ButtonRemoveFromContinueListening":"सुनना जारी रखें से हटाएं","ButtonRemoveSeriesFromContinueSeries":"इस सीरीज को कंटिन्यू सीरीज से हटा दें","ButtonReset":"रीसेट करें","ButtonRestore":"पुनर्स्थापित करें","ButtonSave":"सहेजें","ButtonSaveAndClose":"सहेजें और बंद करें","ButtonSaveTracklist":"ट्रैक सूची सहेजें","ButtonScan":"स्कैन करें","ButtonScanLibrary":"पुस्तकालय स्कैन करें","ButtonSearch":"खोजें","ButtonSelectFolderPath":"फ़ोल्डर का पथ चुनें","ButtonSeries":"सीरीज","ButtonSetChaptersFromTracks":"ट्रैक्स से अध्याय बनाएं","ButtonShiftTimes":"समय खिसकाए","ButtonShow":"दिखाएं","ButtonStartM4BEncode":"M4B एन्कोडिंग शुरू करें","ButtonStartMetadataEmbed":"मेटाडेटा एम्बेडिंग शुरू करें","ButtonSubmit":"जमा करें","ButtonUpload":"अपलोड करें","ButtonUploadBackup":"बैकअप अपलोड करें","ButtonUploadCover":"कवर अपलोड करें","ButtonUploadOPMLFile":"OPML फ़ाइल अपलोड करें","ButtonUserDelete":"उपयोगकर्ता {0} को हटाएं","ButtonUserEdit":"उपयोगकर्ता {0} को संपादित करें","ButtonViewAll":"सभी को देखें","ButtonYes":"हाँ","HeaderAccount":"खाता","HeaderAdvanced":"विकसित","HeaderAppriseNotificationSettings":"Apprise अधिसूचना सेटिंग्स","LabelBackupsMaxBackupSize":"Maximum backup size (in GB)","LabelFontFamily":"फुहारा परिवार","LabelPodcastSearchRegion":"पॉडकास्ट खोज क्षेत्र","LabelSettingsPreferMatchedMetadataHelp":"Matched data will overide item details when using Quick Match. By default Quick Match will only fill in missing details.","MessageBookshelfNoResultsForFilter":"No Results for filter \\"{0}: {1}\\"","NoteChangeRootPassword":"रूट user is the only user that can have an empty password","ToastSendEbookToDeviceFailed":"Failed to Send Ebook to device"}')}}]);