(window.webpackJsonp=window.webpackJsonp||[]).push([[109],{707:function(t,e,n){"use strict";n.r(e);n(14),n(40);var o={props:{value:Boolean,persistent:{type:Boolean,default:!0},width:{type:[String,Number],default:500},height:{type:[String,Number],default:"unset"}},data:function(){return{el:null,content:null}},watch:{show:function(t){t?this.setShow():this.setHide()}},computed:{show:{get:function(){return this.value},set:function(t){this.$emit("input",t)}},modalHeight:function(){return"string"==typeof this.height?this.height:this.height+"px"},modalWidth:function(){return"string"==typeof this.width?this.width:this.width+"px"}},methods:{setShow:function(){var t=this;document.body.appendChild(this.el),setTimeout((function(){t.content.style.transform="scale(1)"}),10)},setHide:function(){this.content.style.transform="scale(0)",this.el.remove()}},mounted:function(){this.el=this.$refs.wrapper,this.content=this.$refs.content,this.content.style.transform="scale(0)",this.content.style.transition="transform 0.25s cubic-bezier(0.16, 1, 0.3, 1)",this.el.style.opacity=1,this.el.remove()}},l=n(4),component=Object(l.a)(o,(function(){var t=this,e=t._self._c;return e("div",{ref:"wrapper",staticClass:"modal modal-bg w-full h-full fixed top-0 left-0 bg-primary/75 flex items-center justify-center z-40 opacity-0"},[e("div",{staticClass:"absolute top-0 left-0 right-0 w-full h-36 bg-linear-to-t from-transparent via-black-500 to-black-700 opacity-90 pointer-events-none"}),t._v(" "),e("div",{ref:"content",staticClass:"relative text-white",staticStyle:{"min-width":"400px","min-height":"200px"},style:{height:t.modalHeight,width:t.modalWidth}},[t._t("default")],2)])}),[],!1,null,null,null);e.default=component.exports}}]);