(window.webpackJsonp=window.webpackJsonp||[]).push([[135,160],{678:function(t,e,l){"use strict";l.r(e);var n={props:{type:{type:String,default:"full"}},data:function(){return{}},computed:{},methods:{},mounted:function(){}},r=l(4),component=Object(r.a)(n,(function(){var t=this,e=t._self._c;return e("div",["bonus"==t.type?[e("ui-tooltip",{attrs:{text:"Bonus",direction:"top"}},[e("span",{staticClass:"material-symbols ml-1",staticStyle:{"font-size":"0.8rem"}},[t._v("local_play")])])]:t._e(),t._v(" "),"trailer"==t.type?[e("ui-tooltip",{attrs:{text:"Trailer",direction:"top"}},[e("span",{staticClass:"material-symbols ml-1",staticStyle:{"font-size":"0.8rem"}},[t._v("local_movies")])])]:t._e()],2)}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{UiTooltip:l(76).default})},875:function(t,e,l){"use strict";l.r(e);var n={props:{queue:{type:Array,default:function(){return[]}},libraryItemId:String},data:function(){return{}},computed:{},methods:{},mounted:function(){}},r=l(4),component=Object(r.a)(n,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"w-full my-2"},[e("div",{staticClass:"w-full bg-primary px-4 md:px-6 py-2 flex items-center"},[e("p",{staticClass:"pr-2 md:pr-4"},[t._v(t._s(t.$strings.HeaderDownloadQueue))]),t._v(" "),e("div",{staticClass:"h-5 md:h-7 w-5 md:w-7 rounded-full bg-white/10 flex items-center justify-center"},[e("span",{staticClass:"text-sm font-mono"},[t._v(t._s(t.queue.length))])])]),t._v(" "),e("transition",{attrs:{name:"slide"}},[e("div",{staticClass:"w-full"},[e("table",{staticClass:"text-sm tracksTable"},[e("tr",[e("th",{staticClass:"text-left px-4 min-w-48"},[t._v(t._s(t.$strings.LabelPodcast))]),t._v(" "),e("th",{staticClass:"text-left w-32 min-w-32"},[t._v(t._s(t.$strings.LabelEpisode))]),t._v(" "),e("th",{staticClass:"text-left px-4"},[t._v(t._s(t.$strings.LabelEpisodeTitle))]),t._v(" "),e("th",{staticClass:"text-left px-4 w-48"},[t._v(t._s(t.$strings.LabelPubDate))])]),t._v(" "),t._l(t.queue,(function(l){return[e("tr",{key:l.id},[e("td",{staticClass:"px-4"},[e("div",{staticClass:"flex items-center"},[e("nuxt-link",{staticClass:"text-sm text-gray-200 hover:underline",attrs:{to:"/item/".concat(l.libraryItemId)}},[t._v(t._s(l.podcastTitle))]),t._v(" "),l.podcastExplicit?e("widgets-explicit-indicator"):t._e()],1)]),t._v(" "),e("td",[e("div",{staticClass:"flex items-center"},[l.season?e("div",[t._v(t._s(l.season)+"x")]):t._e(),t._v(" "),l.episode?e("div",[t._v(t._s(l.episode))]):t._e(),t._v(" "),e("widgets-podcast-type-indicator",{attrs:{type:l.episodeType}})],1)]),t._v(" "),e("td",{staticClass:"px-4",attrs:{dir:"auto"}},[t._v("\n              "+t._s(l.episodeDisplayTitle)+"\n            ")]),t._v(" "),e("td",{staticClass:"text-xs"},[e("div",{staticClass:"flex items-center"},[e("p",[t._v(t._s(t.$dateDistanceFromNow(l.publishedAt)))])])])])]}))],2)])])],1)}),[],!1,null,null,null);e.default=component.exports;installComponents(component,{WidgetsExplicitIndicator:l(354).default,WidgetsPodcastTypeIndicator:l(678).default})}}]);