(window.webpackJsonp=window.webpackJsonp||[]).push([[138],{728:function(t,e,n){"use strict";n.r(e);n(15);var l={props:{title:String,files:{type:Array,default:function(){return[]}}},data:function(){return{expand:!1}},computed:{},methods:{clickBar:function(){this.expand=!this.expand}},mounted:function(){}},r=n(4),component=Object(r.a)(l,(function(){var t=this,e=t._self._c;return e("div",{staticClass:"w-full my-4"},[e("div",{staticClass:"w-full bg-primary px-6 py-1 flex items-center cursor-pointer",on:{click:function(e){return e.stopPropagation(),t.clickBar.apply(null,arguments)}}},[e("p",{staticClass:"pr-4"},[t._v(t._s(t.title))]),t._v(" "),e("span",{staticClass:"bg-black-400 rounded-xl py-0.5 px-2 text-sm font-mono"},[t._v(t._s(t.files.length))]),t._v(" "),e("div",{staticClass:"grow"}),t._v(" "),e("div",{staticClass:"cursor-pointer h-9 w-9 rounded-full hover:bg-black-400 flex justify-center items-center duration-500",class:t.expand?"transform rotate-180":""},[e("span",{staticClass:"material-symbols text-3xl"},[t._v("expand_more")])])]),t._v(" "),e("transition",{attrs:{name:"slide"}},[e("div",{directives:[{name:"show",rawName:"v-show",value:t.expand,expression:"expand"}],staticClass:"w-full"},[e("table",{staticClass:"text-sm tracksTable"},[e("tr",[e("th",{staticClass:"text-left"},[t._v(t._s(t.$strings.LabelFilename))]),t._v(" "),e("th",{staticClass:"text-left"},[t._v(t._s(t.$strings.LabelSize))]),t._v(" "),e("th",{staticClass:"text-left"},[t._v(t._s(t.$strings.LabelType))])]),t._v(" "),t._l(t.files,(function(n){return[e("tr",{key:n.path},[e("td",{staticClass:"pl-2"},[t._v("\n              "+t._s(n.name)+"\n            ")]),t._v(" "),e("td",{staticClass:"font-mono"},[t._v("\n              "+t._s(t.$bytesPretty(n.size))+"\n            ")]),t._v(" "),e("td",[t._v("\n              "+t._s(n.filetype)+"\n            ")])])]}))],2)])])],1)}),[],!1,null,null,null);e.default=component.exports}}]);