(window.webpackJsonp=window.webpackJsonp||[]).push([[151],{876:function(t,n,e){"use strict";e.r(n);var l={data:function(){return{}},computed:{},methods:{},mounted:function(){}},o=e(4),component=Object(o.a)(l,(function(){var t=this,n=t._self._c;return n("ui-tooltip",{staticClass:"inline-flex",attrs:{text:t.$strings.LabelAlreadyInYourLibrary,direction:"top"}},[n("span",{staticClass:"material-symbols ml-1 text-sm text-success"},[t._v("check_circle")])])}),[],!1,null,null,null);n.default=component.exports;installComponents(component,{UiTooltip:e(76).default})}}]);